/***********************************************************************
    PS3 USB Host
    Copyright (C) 2013 Cross Product Creations

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*************************************************************************/

#ifndef __PS3_SPI_H_
#define __PS3_SPI_H_

#include <GenericTypeDefs.h>
#include "outputs.h"
#include <p24Fxxxx.h>
#include "uart2.h"

#define SPI_INPUT_PACKET_LENGTH 25

typedef struct __PACKED _SPI_INPUT_PACKET{
    union{
        struct{
            //override feedback
            unsigned overrideLED:1; //byte 0
            unsigned overrideRumbleL:1;
            unsigned overrideRumbleH:1;
            unsigned :5;
            unsigned char LEDValue; //byte 1
            unsigned char RumbleL;
            unsigned char RumbleH; //byte3

            //override outputs
            unsigned overrideSelect:1; //byte4
            unsigned overrideLStickPress:1;
            unsigned overrideRStickPress:1;
            unsigned overrideStart:1;
            unsigned overrideDPadUp:1;
            unsigned overrideDPadRight:1;
            unsigned overrideDPadDown:1;
            unsigned overrideDPadLeft:1;
            unsigned overrideLTriggerPress:1; //byte5
            unsigned overrideRTriggerPress:1;
            unsigned overrideLBumper:1;
            unsigned overrideRBumper:1;
            unsigned overrideTriangle:1;
            unsigned overrideCircle:1;
            unsigned overrideCross:1;
            unsigned overrideSquare:1;
            unsigned overridePsButton:1; //byte6
            unsigned overrideLStickX:1;
            unsigned overrideLStickY:1;
            unsigned overrideRStickX:1;
            unsigned overrideRStickY:1;
            unsigned overrideLTriggerAnalog:1;
            unsigned overrideRTriggerAnalog:1;
            unsigned overrideAccelX:1;
            unsigned overrideAccelY:1; //byte 7
            unsigned overrideAccelZ:1;
            unsigned overrideGyroZ:1;
            unsigned :5; //padding for byte 7
            unsigned select:1; //byte8
            unsigned lStickPress:1;
            unsigned rStickPress:1;
            unsigned start:1;
            unsigned dPadUp:1;
            unsigned dPadRight:1;
            unsigned dPadDown:1;
            unsigned dPadLeft:1;
            unsigned lTriggerPress:1; //byte 9
            unsigned rTriggerPress:1;
            unsigned lBumper:1;
            unsigned rBumper:1;
            unsigned triangle:1;
            unsigned circle:1;
            unsigned cross:1;
            unsigned square:1;
            unsigned psButton:1; //byte 10
            unsigned :7; //padding for byte 10
            unsigned char lStickX; //byte 11
            unsigned char lStickY;
            unsigned char rStickX;
            unsigned char rStickY;
            unsigned char lTriggerAnalog;
            unsigned char rTriggerAnalog;
            unsigned char accelX_1;
            unsigned char accelX_0;
            unsigned char accelY_1;
            unsigned char accelY_0; //byte 20
            unsigned char accelZ_1;
            unsigned char accelZ_0;
            unsigned char gyroZ_1;
            unsigned char gyroZ_0; //byte 24
        };
        BYTE array[24];
    };
}SPI_INPUT_PACKET;



void initSPI(void);
SPI_INPUT_PACKET* getSpiInputPacket(void);

#endif
