/***********************************************************************
    PS3 USB Host
    Copyright (C) 2013 Cross Product Creations

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*************************************************************************/

#ifndef __PS3_CONTROLLER_H_
#define __PS3_CONTROLLER_H_

#include "GenericTypeDefs.h"

#define PS3_REPORT_LENGTH 49

typedef struct __PACKED _CONTROLLER_IN
{
    //header data in BT mode
    unsigned int l2capHandle;
    unsigned int l2capLength;
    unsigned int frameLength;
    unsigned int channelID;
    unsigned char header;

    //start of packet in both USB and BT modes
    unsigned char packetType;
    unsigned char empty; //not sure what this byte does, but if it's 0xFF, the packet is empty garbage
    unsigned select:1;
    unsigned lStickPress:1;
    unsigned rStickPress:1;
    unsigned start:1;
    unsigned dPadUp:1;
    unsigned dPadRight:1;
    unsigned dPadDown:1;
    unsigned dPadLeft:1;
    unsigned lTriggerPress:1;
    unsigned rTriggerPress:1;
    unsigned lBumper:1;
    unsigned rBumper:1;
    unsigned triangle:1;
    unsigned circle:1;
    unsigned cross:1;
    unsigned square:1;
    unsigned psButton:1;
    unsigned :7;
    unsigned :8;
    unsigned char lStickX;
    unsigned char lStickY;
    unsigned char rStickX;
    unsigned char rStickY;
    unsigned :8; //padding reserved for using 10-bit stick mode
    unsigned :8; //padding reserved for using 10-bit stick mode
    unsigned :8; //padding reserved for using 10-bit stick mode
    unsigned :8; //padding reserved for using 10-bit stick mode
    unsigned char dPadUpPressure;
    unsigned char dPadRightPressure;
    unsigned char dPadDownPressure;
    unsigned char dPadLeftPressure;
    unsigned char lTriggerAnalog;
    unsigned char rTriggerAnalog;
    unsigned char lBumperPressure;
    unsigned char rBumperPressure;
    unsigned char trianglePressure;
    unsigned char circlePressure;
    unsigned char crossPressure;
    unsigned char squarePressure;
    unsigned :8;
    unsigned :8;
    unsigned :8;
    unsigned char usbCable; ///0x02 = plugged, 0x03 = unplugged
    unsigned char battery; //0x05 = fully charged, 0x01 = dead
    unsigned char rumbleStatus; //0x10 = usb, rumble on. 0x12 = usb, rumble off. 0x14 = bluetooth, rumble on. 0x16 bluetooth, rumble off.
    unsigned :8;  //not sure about this one, turns to 0xFF when RSSI is first updated.
    unsigned char RSSI;  //looks like a signal strength indicator, not sure though.  Updated every 10 secs.
    unsigned :8;
    unsigned :8;
    unsigned :8;
    unsigned :8;
    unsigned :8;
    unsigned :8;
    unsigned :8;
    unsigned char accelX_1;
    unsigned char accelX_0;
    unsigned char accelY_1;
    unsigned char accelY_0;
    unsigned char accelZ_1;
    unsigned char accelZ_0;
    unsigned char gyroZ_1;
    unsigned char gyroZ_0;
}CONTROLLER_IN;

/*
 *  outputReportBT[11]=0x10;// duration for fast rumble, about 20ms per count. 0xFF is persistent on.
 *  outputReportBT[12]=0x01;// Bit 0 is on/off for fast rumble?
 *  outputReportBT[13]=0x20;// duration for slow rumble
 *  outputReportBT[14]=0x00;// power for slow rumble
 *  outputReportBT[19]=0x00;// LED
 *  LEDs:
 *      LED1/2/3/4 = 0x02/0x04/0x08/0x10
 */
typedef struct __PACKED _CONTROLLER_OUTPUT
{
    unsigned char alternateMode; //enables alternate modes. Bit7:enable aternate mode, Bit0:10-bits for all inputs, Bit1: 10-bit sticks, Bit2: 8-bits for accels/gyros
    unsigned char rumbleHDuration; ///outputReportBT[11]
    unsigned rumbleHOn:1; //outputReportBT[12]
    unsigned :7;
    unsigned char rumbleLDuration; //outputReportBT[13]
    unsigned char rumbleLPower; //outputReportBT[14]
    unsigned char gyroEnable; //seems like this value must be >=6 for stable gyro readings
    unsigned char gyroCalibration; //affects the bias voltage of the gyro
    unsigned char byte7; //
    unsigned char byte8; //
    unsigned char LED;
}CONTROLLER_OUTPUT;


typedef union _BUTTONS
{
    struct 
    {
        unsigned int select;
        unsigned int lStickPress;
        unsigned int rStickPress;
        unsigned int start;
        unsigned int dPadUp;
        unsigned int dPadRight;
        unsigned int dPadDown;
        unsigned int dPadLeft;
        unsigned int lTriggerPress;
        unsigned int rTriggerPress;
        unsigned int lBumper;
        unsigned int rBumper;
        unsigned int triangle;
        unsigned int circle;
        unsigned int cross;
        unsigned int square;
        unsigned int psButton;
    };
    unsigned int array[17];
} BUTTONS;


typedef union _ANALOGS
{
    struct
    {
        unsigned int lStickX;
        unsigned int lStickY;
        unsigned int rStickX;
        unsigned int rStickY;
        unsigned int lTriggerAnalog;
        unsigned int rTriggerAnalog;
        unsigned int accelX;
        unsigned int accelY;
        unsigned int accelZ;
        unsigned int gyroZ;
    };
    unsigned int array[10];
} ANALOGS;

typedef struct _CONTROLLER_STATE_BUFFER
{
    //buttons
    BUTTONS Buttons;
    ANALOGS Analogs;
    unsigned char usbCable; ///0x02 = plugged, 0x03 = unplugged
    unsigned char battery; //0x05 = fully charged, 0x01 = dead
    unsigned char rumbleStatus; //0x10 = usb, rumble on. 0x12 = usb, rumble off. 0x14 = bluetooth, rumble on. 0x16 bluetooth, rumble off.
}CONTROLLER_STATE_BUFFER;

void InitHIDOutputReport(void);

//////////////////////////////UPDATE BUTTONS/////////////////////////////
void UpdateCurrentStateBuffer(CONTROLLER_IN *ControllerIn);
void SpiInputOverride(void);
void UpdateNewPress(void);
//void UpdatePS3ButtonDuration(void);

//////////////////////////////GET BUTTON INFO//////////////////////////////////

BUTTONS* GetButtonPress(void); //For checking if a button is pressed

BUTTONS* GetPressDuration(void); //For checking how long a button is pressed

BUTTONS* GetNewPress(void); //For checking if a button was pressed since the last poll

ANALOGS* GetAnalogs(void); //For checking the analog inputs

BYTE* GetPS3Report(void);

CONTROLLER_OUTPUT* GetOutputPacket(void);

////////////////////////SET LED FUNCTIONS//////////////////////////////
void SetLEDBin(unsigned char value);
void SetLED1(BOOL value);
void SetLED2(BOOL value);
void SetLED3(BOOL value);
void SetLED4(BOOL value);

////////////////////////Vibe SET FUNCTIONS/////////////////////////
void SetRumbleH(unsigned char level);
void SetRumbleL(unsigned char level);

void UpdateBatteryDisplay(CONTROLLER_IN* Controller);
void UpdateRumbleFeedback(unsigned int duration);

///////////////////////MISC FUNCTIONS///////////////////////////////
unsigned int GetIdleTimer(void);
void ResetIdleTimer(void);

#endif

