/***********************************************************************
    PS3 USB Host
    Copyright (C) 2013 Cross Product Creations

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*************************************************************************/

#include "outputs.h"
#include "main.h"
#include "HardwareProfile.h"
#include "uart2.h"
#include "DEE Emulation 16-bit/DEE Emulation 16-bit.h"
#include <stdlib.h>
//#include <math.h>

/********************************************
 * Index: Input
 * //////////////SERVOS///////////////
 * 0: Left Stick x-axis
 * 1: Left Stick y-axis
 * 2: Right Stick x-axis
 * 3: Right Stick y-axis
 * 4: Left Trigger Analog
 * 5: Right Trigger Analog
 * 6: X-accel
 * 7: Y-accel
 * 8: Z-accel
 * 9: Z-gyro
 * //////////////BUTTONS///////////////
 * 10: select
 * 11: lStickPress
 * 12: rStickPress
 * 13: start
 * 14: dPadUp
 * 15: dPadRight
 * 16: dPadDown
 * 17: dPadLeft
 * 18: lTriggerPress
 * 19: rTriggerPress
 * 20: lBumper
 * 21: rBumper
 * 22: triangle
 * 23: circle
 * 24: cross
 * 25: square
 * 26: psButton
**********************************************/
#define L_STICK_X 0
#define L_STICK_Y 1
#define R_STICK_X 2
#define R_STICK_Y 3
#define L_TRIG 4
#define R_TRIG 5
#define ACCEL_X 6
#define ACCEL_Y 7
#define ACCEL_Z 8
#define GYRO 9
#define PWM10_OUT _LATB8

///////////////////////PWM REGISTERS//////////////////
static unsigned int OC10R, OC10RS, OC10CON1;
static volatile unsigned int* const OCxR_REG_PTRS[10] = {&OC1R,&OC2R,&OC3R,&OC4R,&OC5R,&OC6R,&OC7R,&OC8R,&OC9R,&OC10R};
static volatile unsigned int* const OCxRS_REG_PTRS[10] = {&OC1RS,&OC2RS,&OC3RS,&OC4RS,&OC5RS,&OC6RS,&OC7RS,&OC8RS,&OC9RS,&OC10RS};
static volatile unsigned int* const OCxCON1_REG_PTRS[10] = {&OC1CON1,&OC2CON1,&OC3CON1,&OC4CON1,&OC5CON1,&OC6CON1,&OC7CON1,&OC8CON1,&OC9CON1,&OC10CON1};

///////////////CONFIG SETTINGS/////////////////////////
static SERVO_SETTINGS ServoSettings[NUM_SERVOS];
static signed long int servoState[NUM_SERVOS] = {300000,300000,300000,300000,300000,300000,300000,300000,300000,300000};
static BUTTON_SETTINGS ButtonSettings[NUM_BUTTONS];
static BOOL triggerLink = FALSE;
static BUTTONS buttonToggles;
static unsigned int spiBusMode = 0;
static unsigned int disconnectToggleReset = 0;
static unsigned int idleTimeout = 60000;
static BOOL lStickDiffMix = FALSE;
static BOOL rStickDiffMix = FALSE;
static enum {
    MECANUM_OFF = 0,
    MECANUM_ON_NO_REMAP = 1,
    MECANUM_REMAP_4 = 2,
    MECANUM_REMAP_5 = 3,
    MECANUM_REMAP_6 = 4,
    MECANUM_REMAP_7 = 5,
    MECANUM_REMAP_8 = 6,
    MECANUM_REMAP_9 = 7,
} mecanumSteering = MECANUM_OFF;


//////////////PUBLIC FUNCTIONS///////////////////////////
void InitOutputs(void){
    unsigned int i;

    //Set initial Servo period, time base Timer 2
    T2CON = 0; //clear state
    T2CONbits.TSIDL = 0; //Continue in Idle
    T2CONbits.TGATE = 0; //No gate accumulation
    T2CONbits.TCKPS = 0b01; //Prescale 1:8
    T2CONbits.T32 = 0; //16-bit mode
    T2CONbits.TCS = 0; //Clock Source: Internal Clock (Fosc/2)
    TMR2 = 0;
    T2CONbits.TON = 1; //Timer2 On
    //***
    
    for (i = 0; i<NUM_SERVOS; i++)
    {
        *OCxCON1_REG_PTRS[i] = 0b110;
        *OCxRS_REG_PTRS[i] = 40000;
        *OCxR_REG_PTRS[i] = 3000;
    }

    OC1CON2 = 0x003F;//use own OCxRS register for sync
    OC2CON2 = 0x003F;//use own OCxRS register for sync
    OC3CON2 = 0x003F;//use own OCxRS register for sync
    OC4CON2 = 0x003F;//use own OCxRS register for sync
    OC5CON2 = 0x003F;//use own OCxRS register for sync
    OC6CON2 = 0x003F;//use own OCxRS register for sync
    OC7CON2 = 0x003F;//use own OCxRS register for sync
    OC8CON2 = 0x003F;//use own OCxRS register for sync
    OC9CON2 = 0x003F;//use own OCxRS register for sync
    
    //Stagger the servo signal turn-on sequence to reduce the inrush current
    RPOR10bits.RP21R = 18; //Map PWM 1 Output to RP21
    OC1CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR13bits.RP26R = 19; //Map PWM 2 Output to RP26
    OC2CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR9bits.RP19R = 20; //Map PWM 3 Output to RP19
    OC3CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR13bits.RP27R = 21; //Map PWM 4 Output to RP27
    OC4CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR9bits.RP18R = 22; //Map PWM 5 Output to RP18
    OC5CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR14bits.RP28R = 23; //Map PWM 6 Output to RP28
    OC6CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR6bits.RP13R = 24; //Map PWM 7 Output to RP13
    OC7CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR0bits.RP1R = 25; //Map PWM 8 Output to RP1
    OC8CON2bits.OCTRIS = 0;
    DelayMs(100);
    RPOR0bits.RP0R = 35; //Map PWM 9 Output to RP0
    OC9CON2bits.OCTRIS = 0;

/////////////Use Timer3 for PWM 10
    T3CON = 0; //clear state
    T3CONbits.TSIDL = 0; //Continue in Idle
    T3CONbits.TGATE = 0; //No gate accumulation
    T3CONbits.TCKPS = 0b01; //Prescale 1:8
    T3CONbits.TCS = 0; //Clock Source: Internal Clock (Fosc/2)
    TMR3 = 0;
    T3CONbits.TON = 1;
        //enable timer4 interrupt
    _T3IE = 1;
    _T3IP = 7;//high priority interrupt
//////////////////////////////////
}


/*******************************************************************************
Function: LoadDefaultSettings(BOOL loadAll, unsigned int index)

Precondition:
 None

Overview:
 load default settings

Input:
 loadAll:load all the default settings
 index: load the default settings for the channel if loadAll is False.

Output:
 None

*******************************************************************************/
void LoadDefaultSettings(BOOL loadAll, unsigned int index)
{
    unsigned int i = 0;
    if (loadAll == FALSE)
    {
        i = index;
        UART2PrintString("Loading default settings for ");
        PrintIndexLabel(i);
    }
    else
    {
        UART2PrintString("Loading all default settings...");
        triggerLink = FALSE;
        spiBusMode = 0;
        disconnectToggleReset = 0;
        idleTimeout = 60000;
        lStickDiffMix = FALSE;
        rStickDiffMix = FALSE;
        mecanumSteering = MECANUM_OFF;
    }
    do
    {
        if (i < NUM_SERVOS)
        {
            ServoSettings[i].offset = 0;
            ServoSettings[i].range = 125;
            ServoSettings[i].deadband = 0;
            ServoSettings[i].sensitivity = 10;
            ServoSettings[i].direction = FORWARD;
            ServoSettings[i].absOrRel = ABSOLUTE;
            ServoSettings[i].servoOrMotor = SERVO;
            ServoSettings[i].servoPeriodMs = 20;
            ServoSettings[i].pwmFrequency = 10;
            ServoSettings[i].trim = 0;
            ServoSettings[i].zeroPosition = 128;
            ServoSettings[i].holdRecall = OFF;
            ServoSettings[i].servoRecallValue = 3000;

            
            if ( i == ACCEL_X || i == ACCEL_Y || i == ACCEL_Z || i == GYRO)
            {
                ServoSettings[i].zeroPosition = 512;
            }
            else if (i == L_TRIG || i == R_TRIG)
            {
                ServoSettings[i].zeroPosition = 0;
            }
            else
            {
                ServoSettings[i].zeroPosition = 128;
            }

            DataEEWrite(3000, EEPROM_SERVO_HOME + i);
        }
        else
        {
            ButtonSettings[i-NUM_SERVOS].invert = FALSE;
            ButtonSettings[i-NUM_SERVOS].outputMode = PUSHBUTTON;
            ButtonSettings[i-NUM_SERVOS].pulseWidth = 5;
            ButtonSettings[i-NUM_SERVOS].autofirePeriod = 20;
        }
        i++;
    } while(i < (NUM_SERVOS+NUM_BUTTONS) && loadAll == TRUE);
    UART2PrintString("Done\n\r");
}


/*******************************************************************************
Function: SaveLoadSubroutine( BOOL save)

Precondition:
 None

Overview:
  Save or load settings.

Input:
  if save==1, then save, if save==0, then load

Output:
 None

*******************************************************************************/

void SaveLoadSubroutine( BOOL save_switch )
{
    unsigned int i = 0;
    unsigned int j = 0;
    unsigned int memory_index = 0;

    //servo settings
    for (i = 0; i <NUM_SERVOS; i++)
    {
        for (j = 0; j<SERVO_SETTINGS_LENGTH; j++)
        {
            if (save_switch) DataEEWrite(ServoSettings[i].array[j], EEPROM_CONFIG_SETTINGS_START + memory_index);
            else ServoSettings[i].array[j] = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
            memory_index++;
        }
    }

    //button settings
    for (i = 0; i<NUM_BUTTONS; i++)
    {
        for (j = 0; j<4; j++)
        {
            if (save_switch) DataEEWrite(ButtonSettings[i].array[j], EEPROM_CONFIG_SETTINGS_START + memory_index);
            else ButtonSettings[i].array[j] = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
            memory_index++;
        }
    }

    //universal settings
    if (save_switch) DataEEWrite(triggerLink, EEPROM_CONFIG_SETTINGS_START + memory_index);
    else triggerLink = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
    memory_index++;

    if (save_switch) DataEEWrite(spiBusMode, EEPROM_CONFIG_SETTINGS_START + memory_index);
    else spiBusMode = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
    memory_index++;

    if (save_switch) DataEEWrite(disconnectToggleReset, EEPROM_CONFIG_SETTINGS_START + memory_index);
    else disconnectToggleReset = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
    memory_index++;

    if (save_switch) DataEEWrite(idleTimeout, EEPROM_CONFIG_SETTINGS_START + memory_index);
    else idleTimeout = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
    memory_index++;

    if (save_switch) DataEEWrite(lStickDiffMix, EEPROM_CONFIG_SETTINGS_START + memory_index);
    else lStickDiffMix = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
    memory_index++;

    if (save_switch) DataEEWrite(rStickDiffMix, EEPROM_CONFIG_SETTINGS_START + memory_index);
    else rStickDiffMix = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
    memory_index++;

    if (save_switch) DataEEWrite(mecanumSteering, EEPROM_CONFIG_SETTINGS_START + memory_index);
    else mecanumSteering = DataEERead(EEPROM_CONFIG_SETTINGS_START + memory_index);
    memory_index++;

    UART2PutDec(memory_index);
    UART2PrintString(" EEPROM words used.\n\r");

    //when adding settings here, don't forget to update the default load settings function above
}


/*******************************************************************************
Function: LoadSavedSettings()

Precondition:
 None

Overview:
 Wrapper for SaveLoadSettings, loads previously saved settings

Input:
  None

Output:
 None

*******************************************************************************/
void LoadSavedSettings(void)
{
    UART2PrintString("Loading saved settings...");
    SaveLoadSubroutine(0);
    UART2PrintString("Done\n\r");
}

/*******************************************************************************
Function: SaveSettings()

Precondition:
 None

Overview:
 Wrapper for SaveLoadSettings, saves settings

Input:
  None

Output:
 None

*******************************************************************************/
void SaveSettings(void)
{

    UART2PrintString("Saving Settings...");
    SaveLoadSubroutine(1);
    UART2PrintString("Done.\n\r");
}

//Return servo settings
SERVO_SETTINGS* GetServoSettings (void)
{
    return ServoSettings;
}

BUTTON_SETTINGS* GetButtonSettings (void)
{
    return ButtonSettings;
}

unsigned int GetTimeoutSetting(void)
{
    return idleTimeout;
}

/*******************************************************************************
Function: ConfigOutput()

Precondition:
 None

Overview:
    This routine allows the user to modify output settings

Input: 
 configFlag == TRUE will continue to run the routine.
 configFlag == FALSE will exit the routine.

Output:
  0 = save and exited;
  1 = stay in config mode;

*******************************************************************************/

BOOL ConfigOutput (BOOL configFlag)
{
    static BOOL configCmdReleased = FALSE;
    static unsigned int cycleCounter = 0; //increment each time we enter the function
    static unsigned int index = 0; //index to mark which output we're configuring
    static BOOL settingsSaved;

/********************************************
 * Index: Input
 * //////////////SERVOS///////////////
 * 0: Left Stick x-axis
 * 1: Left Stick y-axis
 * 2: Right Stick x-axis
 * 3: Right Stick y-axis
 * 4: Left Trigger Analog
 * 5: Right Trigger Analog
 * 6: X-accel
 * 7: Y-accel
 * 8: Z-accel
 * 9: Z-gyro
 * //////////////BUTTONS///////////////
 * 10: select
 * 11: lStickPress
 * 12: rStickPress
 * 13: start
 * 14: dPadUp
 * 15: dPadRight
 * 16: dPadDown
 * 17: dPadLeft
 * 18: lTriggerPress
 * 19: rTriggerPress
 * 20: lBumper
 * 21: rBumper
 * 22: triangle
 * 23: circle
 * 24: cross
 * 25: square
 * 26: psButton
**********************************************/

    /////////////////////////////////////////////////////////////////////////////////
    //wait for user to release the "enter config" command buttons so it won't start
    //making changes right away.
    if (configCmdReleased==FALSE && GetPressDuration()->select == 0 && GetPressDuration()->psButton == 0)
    {
        //initialize config mode flags
        UART2PrintString("Entering Config Mode:\n\r");
        configCmdReleased = TRUE;
        index = 0;
        cycleCounter = 0;
        PrintIndexLabel(index);
        UpdateButtonOutputs(FALSE, 0); //update the button outputs before we start and reset the "select" and "PS button" outputs.

    }
    else if (configCmdReleased == FALSE) //force user to release buttons so we don't start making changes right away
    {
        UpdateRumbleFeedback(2);
        return TRUE;
    }


    //////////exit when button is released to prevent the Start button output from activating
    //////////immediately upon exit.
    cycleCounter++;

    ////////////////////Commands here apply to both servos and buttons////////////////////////
    if (!GetButtonPress()->rStickPress && !GetButtonPress()->lStickPress)
    {
        //update index
        if (GetNewPress()->rBumper)
        {
            index++;
            index = index%(NUM_SERVOS+NUM_BUTTONS);
            PrintIndexLabel(index);
        }
        if (GetNewPress()->lBumper)
        {
            if (index == 0)
            {
                index = (NUM_SERVOS+NUM_BUTTONS)-1;
            }
            else
            {
                index--;
            }
            index = index%(NUM_SERVOS+NUM_BUTTONS);
            PrintIndexLabel(index);
        }
        SetLEDBin(index);

        //Reset Defaults
        if (GetPressDuration()->select == 100 && GetButtonPress()->start == 0) //check start button to make sure user isn't trying to disconnect
        {
            LoadDefaultSettings(FALSE,index);
            UpdateRumbleFeedback(5);
        }

        if (GetPressDuration()->select == 300 && GetButtonPress()->start == 0) //check start button to make sure user isn't trying to disconnect
        {
            LoadDefaultSettings(TRUE,0);
            UpdateRumbleFeedback(10);
        }

        //Save and exit
        if (GetPressDuration()->start == 100 && GetButtonPress()->select == 0) //check select button to make sure user isn't trying to disconnect
        {
            SaveSettings();
            settingsSaved = TRUE;
            UpdateRumbleFeedback(5);
        }
        else if ((GetButtonPress()->start == 0 && settingsSaved == TRUE) || configFlag == FALSE)
        {
            UART2PrintString("Exiting Config Mode.\n\r");
            //reset flags
            configCmdReleased = FALSE;
            settingsSaved = FALSE;
            return FALSE;
        }
    }
    else if (GetButtonPress()->rStickPress) //if right stick switch is pressed
    {
        //Cycle SPI bus modes;
        if (GetNewPress()->triangle)
        {
            spiBusMode++;
            spiBusMode = spiBusMode%4;
            switch (spiBusMode)
            {
                case 0:
                    UART2PrintString("SPI Bus Mode 0\n\r");
                    SPI1STATbits.SPIEN = 0;
                    SPI1CON1bits.CKP = 0;
                    SPI1CON1bits.CKE = 1;
                    SPI1STATbits.SPIEN = 1;
                    break;

                case 1:
                    UART2PrintString("SPI Bus Mode 1\n\r");
                    SPI1STATbits.SPIEN = 0;
                    SPI1CON1bits.CKP = 0;
                    SPI1CON1bits.CKE = 0;
                    SPI1STATbits.SPIEN = 1;
                    break;

                case 2:
                    UART2PrintString("SPI Bus Mode 2\n\r");
                    SPI1STATbits.SPIEN = 0;
                    SPI1CON1bits.CKP = 1;
                    SPI1CON1bits.CKE = 1;
                    SPI1STATbits.SPIEN = 1;
                    break;

                case 3:
                    UART2PrintString("SPI Bus Mode 3\n\r");
                    SPI1STATbits.SPIEN = 0;
                    SPI1CON1bits.CKP = 1;
                    SPI1CON1bits.CKE = 0;
                    SPI1STATbits.SPIEN = 1;
                    break;

                default:
                    UART2PrintString("ERROR SPI BUS MODE\n\r");
                    break;
            }
        }
        if (GetNewPress()->cross)
        {
            if (idleTimeout == 0)
            {
                idleTimeout = 60000;
                UART2PrintString("Idle auto-off set to 10 min.\n\r");
            }
            else
            {
                idleTimeout = 0;
                UART2PrintString("Idle auto-off disabled.\n\r");
            }
        }
        if (GetNewPress()->psButton)
        {
            disconnectToggleReset = !disconnectToggleReset;
            if (disconnectToggleReset) UART2PrintString("Toggled outputs reset on disconnect.\n\r");
            else UART2PrintString("Toggled outputs preserved on disconnect.\n\r");
        }
    }
    else if (GetButtonPress()->lStickPress) //if left stick switch is pressed
    {
    }

    /////////////////SET SERVO SETTINGS///////////////////////
    if (index < NUM_SERVOS)
    {
        //if the RStick button is held down, enable different set of options.
        if (GetButtonPress()->rStickPress)
        {

            //Increase output trim
            if (GetNewPress()->dPadRight || (GetPressDuration()->dPadRight > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].trim < TRIM_MAX)
                {
                    ServoSettings[index].trim++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Trim = ");
                UART2PutDecSInt(ServoSettings[index].trim);
                UART2PrintString("\n\r");
            }

            //Decrease output trim
            if (GetNewPress()->dPadLeft || (GetPressDuration()->dPadLeft > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].trim >TRIM_MIN)
                {
                    ServoSettings[index].trim--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Trim = ");
                UART2PutDecSInt(ServoSettings[index].trim);
                UART2PrintString("\n\r");
            }

            //Increase PWM Signal Period
            if (GetNewPress()->dPadUp || (GetPressDuration()->dPadUp > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].pwmFrequency < PWM_FREQ_MAX_KHZ)
                {
                    ServoSettings[index].pwmFrequency++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("PWM Frequency = ");
                UART2PutDecSInt(ServoSettings[index].pwmFrequency);
                UART2PrintString(" kHz.\n\r");
            }

            //Decrease PWM Period
            if (GetNewPress()->dPadDown || (GetPressDuration()->dPadDown > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].pwmFrequency > PWM_FREQ_MIN_KHZ)
                {
                    ServoSettings[index].pwmFrequency--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("PWM Frequency = ");
                UART2PutDecSInt(ServoSettings[index].pwmFrequency);
                UART2PrintString(" kHz.\n\r");
            }

            //Set servo forward/reverse
            if (GetNewPress()->lBumper)
            {
                ServoSettings[index].direction = !ServoSettings[index].direction;
                if (ServoSettings[index].direction == FORWARD)
                {
                    UART2PrintString("Servo/Motor Forward.\n\r");
                }
                else if (ServoSettings[index].direction == REVERSE)
                {
                    UART2PrintString("Servo/Motor Reverse.\n\r");
                }
            }

            //Set absolute/relative mode
            if (GetNewPress()->rBumper)
            {
                ServoSettings[index].absOrRel  = !ServoSettings[index].absOrRel;
                if (ServoSettings[index].absOrRel == ABSOLUTE)
                {
                    UART2PrintString("Absolute input mode.\n\r");
                }
                else if (ServoSettings[index].absOrRel == RELATIVE)
                {
                    UART2PrintString("Relative input mode.\n\r");
                }
            }

        }//end settings if rstick is held down

        //settings if L-Stick is held down
        else if (GetButtonPress()->lStickPress)
        {
            //Increase Servo Signal Period
            if (GetNewPress()->dPadUp  || (GetPressDuration()->dPadUp > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].servoPeriodMs < SERVO_PERIOD_MS_MAX)
                {
                    ServoSettings[index].servoPeriodMs++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Servo Signal Period = ");
                UART2PutDecSInt(ServoSettings[index].servoPeriodMs);
                UART2PrintString(" ms.\n\r");
            }

            //Decrease Servo Period
            if (GetNewPress()->dPadDown || (GetPressDuration()->dPadDown > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].servoPeriodMs > SERVO_PERIOD_MS_MIN)
                {
                    ServoSettings[index].servoPeriodMs--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Servo Signal Period = ");
                UART2PutDecSInt(ServoSettings[index].servoPeriodMs);
                UART2PrintString(" ms.\n\r");
            }
            
            //SET TRIGGER LINK
            /*linking the trigger linkes the 2 triggers so that one move the servo in a
             * positive direction and the other moves the servo in a negative direction.*/
            if ((index == L_TRIG || index == R_TRIG) && GetNewPress()->cross)
            {
                triggerLink = !triggerLink;
                if (triggerLink == TRUE)
                {
                    UART2PrintString("Trigger Link On.\n\r");
                }
                else
                {
                    UART2PrintString("Trigger Link Off.\n\r");
                }
            }

            //Left Stick only settings
            if (index == L_STICK_X || index == L_STICK_Y)//Left stick settings
            {
                if (GetNewPress()->cross) //Left Stick Differential Mix Enable/Disable
                {
                    lStickDiffMix = !lStickDiffMix;
                    if (lStickDiffMix == TRUE)
                    {
                        UART2PrintString("Left Stick Differential Mix On.\n\r");
                    }
                    else
                    {
                        UART2PrintString("Left Stick Differential Mix Off.\n\r");
                    }
                }
            }
            //Right Stick Settings
            if (index == R_STICK_X || index == R_STICK_Y)
            {
                if (GetNewPress()->cross) //Right Stick Differential Mix Enable/Disable
                {
                    rStickDiffMix = !rStickDiffMix;
                    if (rStickDiffMix == TRUE)
                    {
                        UART2PrintString("Right Stick Differential Mix On.\n\r");
                    }
                    else
                    {
                        UART2PrintString("Right Stick Differential Mix Off.\n\r");
                    }
                }
            }

            //Configurations valid for both sticks
            if (GetNewPress()->square) //mecanum steering
            {
                mecanumSteering++;
                mecanumSteering = mecanumSteering%8;

                switch(mecanumSteering)
                {
                    case MECANUM_OFF:
                        UART2PrintString("Mecanum Steering Mix Off.\n\r");
                        break;
                    case MECANUM_ON_NO_REMAP:
                        UART2PrintString("Mecanum Steering Mix On.\n\r");
                        break;
                    default:
                        UART2PrintString("Mecanum Steering Mix On, remap right stick y-axis to ");
                        PrintIndexLabel(mecanumSteering+2);
                        break;
                }
            }
                
            if (GetNewPress()->circle)
            {
                ServoSettings[index].holdRecall++;
                ServoSettings[index].holdRecall = ServoSettings[index].holdRecall%3;
                switch (ServoSettings[index].holdRecall)
                {
                    case OFF:
                        UART2PrintString("Stick Hold/Recall disabled.\n\r");
                        break;
                    case L_STICK:
                        UART2PrintString("Stick Hold/Recall tied to L-Stick Button.\n\r");
                        break;
                    case R_STICK:
                        UART2PrintString("Stick Hold/Recall tied to R-Stick Button.\n\r");
                        break;
                    default:
                        break;
                }
            }
            
            //Set RC Servo/PWM mode
            if (GetNewPress()->lBumper)
            {
                ServoSettings[index].servoOrMotor = !ServoSettings[index].servoOrMotor;
                if (ServoSettings[index].servoOrMotor == SERVO)
                {
                    UART2PrintString("RC servo mode.\n\r");
                }
                else if (ServoSettings[index].servoOrMotor == MOTOR)
                {
                    UART2PrintString("PWM mode.\n\r");
                }
            }
        }//end settings if L-Stick is held down

        /////////////////////////////////////////////////////
        else //if LSTick or RStick are not held down
        {
        //Increase offset
            if (GetNewPress()->dPadRight || (GetPressDuration()->dPadRight > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].offset < OFFSET_MAX)
                {
                    ServoSettings[index].offset++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Range Offset = ");
                UART2PutDecSInt(ServoSettings[index].offset);
                UART2PrintString("\n\r");
            }

            //Decrease offset
            if (GetNewPress()->dPadLeft ||  (GetPressDuration()->dPadLeft > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].offset > OFFSET_MIN)
                {
                    ServoSettings[index].offset--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Range Offset = ");
                UART2PutDecSInt(ServoSettings[index].offset);
                UART2PrintString("\n\r");
            }

            //Increase range
            if (GetNewPress()->dPadUp || (GetPressDuration()->dPadUp > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].range < RANGE_MAX)
                {
                    ServoSettings[index].range++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Range = ");
                UART2PutDecSInt(ServoSettings[index].range);
                UART2PrintString("\n\r");
            }

            //Decrease range
            if (GetNewPress()->dPadDown ||  (GetPressDuration()->dPadDown > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].range > RANGE_MIN)
                {
                    ServoSettings[index].range--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Range = ");
                UART2PutDecSInt(ServoSettings[index].range);
                UART2PrintString("\n\r");
            }

            //Increase deadband
            if (GetNewPress()->circle || (GetPressDuration()->circle > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].deadband < DEADBAND_MAX)
                {
                    ServoSettings[index].deadband++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Deadband = ");
                UART2PutDecSInt(ServoSettings[index].deadband);
                UART2PrintString("\n\r");
            }

            //Decrease deadband
            if (GetNewPress()->square ||  (GetPressDuration()->square > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].deadband > DEADBAND_MIN)
                {
                    ServoSettings[index].deadband--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Deadband = ");
                UART2PutDecSInt(ServoSettings[index].deadband);
                UART2PrintString("\n\r");
            }
            //Increase sensitivity
            if (GetNewPress()->triangle || (GetPressDuration()->triangle > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].sensitivity < SENSITIVITY_MAX)
                {
                    ServoSettings[index].sensitivity++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Sensitivity = ");
                UART2PutDecSInt(ServoSettings[index].sensitivity);
                UART2PrintString("\n\r");
            }

            //Decrease sensitivity
            if (GetNewPress()->cross ||  (GetPressDuration()->cross > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ServoSettings[index].sensitivity > SENSITIVITY_MIN)
                {
                    ServoSettings[index].sensitivity--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Sensitivity = ");
                UART2PutDecSInt(ServoSettings[index].sensitivity);
                UART2PrintString("\n\r");
            }

            //Recalibration
            if (GetNewPress()->psButton)
            {
                ServoSettings[index].zeroPosition = GetAnalogs()->array[index];
                UART2PrintString("Home Position = ");
                UART2PutDecInt(ServoSettings[index].zeroPosition);
                UART2PrintString("\n\r");
            }
            if (GetPressDuration()->psButton == 100)
            {
                unsigned int i;
                UART2PrintString("Calibrating All Home Positions...");
                for (i = 0; i < NUM_SERVOS; i++)
                {
                        ServoSettings[i].zeroPosition = GetAnalogs()->array[i];
                }
                UART2PrintString("Done.\n\r");
                UpdateRumbleFeedback(10);
            }
        }//end if rStick is not pressed
    }// if i<NUM_SERVOS

    ////////////////SET BUTTONS///////////////////////
    if (index >= NUM_SERVOS && index < (NUM_SERVOS+NUM_BUTTONS))
    {
        unsigned int buttonIndex;
        buttonIndex = index-NUM_SERVOS;

        if (GetButtonPress()->rStickPress){} //if right stick is held
        else if (GetButtonPress()->lStickPress){} //if left stick is held
        else //if neither stick is held
        {
            //set invert
            if (GetNewPress()->cross)
            {
                ButtonSettings[buttonIndex].invert = !ButtonSettings[buttonIndex].invert;
                if (ButtonSettings[buttonIndex].invert == TRUE)
                {
                    UART2PrintString("Output Inverted.\n\r");
                }
                else
                {
                    UART2PrintString("Output Normal.\n\r");
                }
            }

            if (GetNewPress()->triangle)
            {
                //set output style
                switch(ButtonSettings[buttonIndex].outputMode)
                {
                    case PUSHBUTTON:
                        ButtonSettings[buttonIndex].outputMode = TOGGLE;
                        UART2PrintString("Output Toggle.\n\r");
                        break;

                    case TOGGLE:
                        ButtonSettings[buttonIndex].outputMode = SINGLE_SHOT;
                        UART2PrintString("Output Single Shot.\n\r");
                        break;

                    case SINGLE_SHOT:
                        ButtonSettings[buttonIndex].outputMode = AUTOFIRE;
                        UART2PrintString("Output Autofire.\n\r");
                        break;

                    case AUTOFIRE:
                        ButtonSettings[buttonIndex].outputMode = TOGGLE_AUTOFIRE;
                        UART2PrintString("Output Toggle Autofire.\n\r");
                        break;

                    case TOGGLE_AUTOFIRE:
                        ButtonSettings[buttonIndex].outputMode = PUSHBUTTON;
                        UART2PrintString("Output Pushbutton.\n\r");
                        break;

                    default:
                        UART2PrintString("ERROR OUTPUT MODE.\n\r");
                        break;
                }
            }

            //Increase pulse width
            if ((GetPressDuration()->dPadRight > FAST_AUTOCLICK_DELAY)&&\
                (ButtonSettings[buttonIndex].pulseWidth < (PULSE_WIDTH_MAX-FAST_AUTOCLICK_SIZE)))
            {
                ButtonSettings[buttonIndex].pulseWidth += FAST_AUTOCLICK_SIZE;
            }
            if (GetNewPress()->dPadRight || (GetPressDuration()->dPadRight > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ButtonSettings[buttonIndex].pulseWidth < PULSE_WIDTH_MAX)
                {
                    ButtonSettings[buttonIndex].pulseWidth++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("PulseWidth = ");
                UART2PutDecInt(ButtonSettings[buttonIndex].pulseWidth);
                UART2PrintString("0 ms\n\r");
            }

            //Decrease pulse width
            if ((GetPressDuration()->dPadLeft > FAST_AUTOCLICK_DELAY)&&\
                (ButtonSettings[buttonIndex].pulseWidth > (PULSE_WIDTH_MIN + FAST_AUTOCLICK_SIZE)))
            {
                ButtonSettings[buttonIndex].pulseWidth -= FAST_AUTOCLICK_SIZE;
            }
            if (GetNewPress()->dPadLeft ||  (GetPressDuration()->dPadLeft > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ButtonSettings[buttonIndex].pulseWidth > PULSE_WIDTH_MIN)
                {
                    ButtonSettings[buttonIndex].pulseWidth--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("PulseWidth = ");
                UART2PutDecInt (ButtonSettings[buttonIndex].pulseWidth);
                UART2PrintString("0 ms\n\r");
            }

            //Increase Autofire Period
            if ((GetPressDuration()->dPadUp > FAST_AUTOCLICK_DELAY)&&\
                (ButtonSettings[buttonIndex].autofirePeriod < (AUTOFIRE_PERIOD_MAX - FAST_AUTOCLICK_SIZE)))
            {
                ButtonSettings[buttonIndex].autofirePeriod += FAST_AUTOCLICK_SIZE;
            }
            if (GetNewPress()->dPadUp || (GetPressDuration()->dPadUp > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ButtonSettings[buttonIndex].autofirePeriod < AUTOFIRE_PERIOD_MAX)
                {
                    ButtonSettings[buttonIndex].autofirePeriod++;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Autofire Period = ");
                UART2PutDecInt(ButtonSettings[buttonIndex].autofirePeriod);
                UART2PrintString("0 ms\n\r");
            }

            //Decrease Autofire Period
            if ((GetPressDuration()->dPadDown > FAST_AUTOCLICK_DELAY)&&\
                (ButtonSettings[buttonIndex].autofirePeriod > (AUTOFIRE_PERIOD_MIN + FAST_AUTOCLICK_SIZE)))
            {
                ButtonSettings[buttonIndex].autofirePeriod -= FAST_AUTOCLICK_SIZE;
            }
            if (GetNewPress()->dPadDown ||  (GetPressDuration()->dPadDown > AUTOCLICK_DELAY && !(cycleCounter%AUTOCLICK_TIMER)) )
            {
                if (ButtonSettings[buttonIndex].autofirePeriod > AUTOFIRE_PERIOD_MIN)
                {
                    ButtonSettings[buttonIndex].autofirePeriod--;
                }
                else
                {
                    UpdateRumbleFeedback(2);
                }
                UART2PrintString("Autofire Period = ");
                UART2PutDecInt(ButtonSettings[buttonIndex].autofirePeriod);
                UART2PrintString("0 ms\n\r");
            }
        }
    UpdateButtonOutputs(configFlag, buttonIndex);
    }
    return TRUE;
}



/*******************************************************************************
Function: UpdateButtonOutputs(unsigned int configFlag, unsigned int configIndex)

Precondition:
    Must have run all the UpdateNewPress() and UpdatePS3ButtonDuration() routines.

Overview:
    This routine updates the outputs for button presses.

Input: configFlag: informes the routine if configuration mode is active or not.
 * 0: config mode not active
 * 1: config mode active
 * 2: config mode not active, but re-initialize the button toggle states
 *
 If config mode is active, outputs are activated by the right stick
 press button only to avoid conflicts with button functions.

Output: None.

*******************************************************************************/
void UpdateButtonOutputs(unsigned int configFlag, unsigned int configIndex)
{
    //static unsigned int cycleCounter; //increment each time we enter the function
    unsigned int i;
    unsigned int inputIndex;
    unsigned int buttonIndex;
    static BUTTONS buttonOutputs;
    static BUTTONS cycleCounter;


    //Calculate output high or low
    for (buttonIndex = 0; buttonIndex<NUM_BUTTONS; buttonIndex++)
    {
        if (configFlag == 1)
        {
            buttonIndex = NUM_BUTTONS;
            i = configIndex; //only change the output of the button we're configuring
            inputIndex = 2; //use rStickPress for testing the outputs to avoid conflicts with configuration commands;
        }
        else
        {
            i = buttonIndex;
            inputIndex = buttonIndex;
        }
        if (configFlag == 2)
        {
            buttonToggles.array[i] = 0; //reset toggle states to power-on configuration.
        }

        switch (ButtonSettings[i].outputMode)
        {
            case PUSHBUTTON:
                buttonOutputs.array[i] = GetButtonPress()->array[inputIndex];
                break;

            case TOGGLE:
                if ( GetNewPress()->array[inputIndex] )
                {
                    buttonToggles.array[i] = !buttonToggles.array[i];
                }
                buttonOutputs.array[i] = buttonToggles.array[i];
                break;

            case SINGLE_SHOT:
                if ( GetNewPress()->array[inputIndex] )
                {
                    cycleCounter.array[i] = 1;
                    buttonOutputs.array[i] = 1;
                }
                if (cycleCounter.array[i] > 0 && cycleCounter.array[i] <= ButtonSettings[i].pulseWidth)
                {
                    buttonOutputs.array[i] = 1;
                    cycleCounter.array[i]++;
                }
                else
                {
                    cycleCounter.array[i] = 0;
                    buttonOutputs.array[i] = 0;
                }
                break;

            case AUTOFIRE:
                if (GetButtonPress()->array[inputIndex])
                {
                    if ( cycleCounter.array[i]%ButtonSettings[i].autofirePeriod < ButtonSettings[i].pulseWidth)
                    {
                        buttonOutputs.array[i] = 1;
                    }
                    else
                    {
                        buttonOutputs.array[i] = 0;
                    }
                    cycleCounter.array[i]++;
                }
                else
                {
                    buttonOutputs.array[i] = 0;
                    cycleCounter.array[i] = 0;
                }
                break;

            case TOGGLE_AUTOFIRE:
                if ( GetNewPress()->array[inputIndex] )
                {
                    buttonToggles.array[i] = !buttonToggles.array[i];
                }
                if (buttonToggles.array[i])
                {
                    if ( cycleCounter.array[i]%ButtonSettings[i].autofirePeriod < ButtonSettings[i].pulseWidth)
                    {
                        buttonOutputs.array[i] = 1;
                    }
                    else
                    {
                        buttonOutputs.array[i] = 0;
                    }
                    cycleCounter.array[i]++;
                }
                else
                {
                    buttonOutputs.array[i] = 0;
                    cycleCounter.array[i] = 0;
                }
                break;

            default:
                break;
        }//end switch

        //We must explicitly assign a value to the output each time, or else the
        //following code will keep toggling the output if inversion is on.
        if (ButtonSettings[i].invert == TRUE)
        {
            buttonOutputs.array[i] = !buttonOutputs.array[i];
        }
    }//end for loop

    
    //Update the physical addresses associated with the ouput pins
    BUTTON_OUT_0 = buttonOutputs.array[0];
    BUTTON_OUT_1 = buttonOutputs.array[1];
    BUTTON_OUT_2 = buttonOutputs.array[2];
    BUTTON_OUT_3 = buttonOutputs.array[3];
    BUTTON_OUT_4 = buttonOutputs.array[4];
    BUTTON_OUT_5 = buttonOutputs.array[5];
    BUTTON_OUT_6 = buttonOutputs.array[6];
    BUTTON_OUT_7 = buttonOutputs.array[7];
    BUTTON_OUT_8 = buttonOutputs.array[8];
    BUTTON_OUT_9 = buttonOutputs.array[9];
    BUTTON_OUT_10 = buttonOutputs.array[10];
    BUTTON_OUT_11 = buttonOutputs.array[11];
    BUTTON_OUT_12 = buttonOutputs.array[12];
    BUTTON_OUT_13 = buttonOutputs.array[13];
    BUTTON_OUT_14 = buttonOutputs.array[14];
    BUTTON_OUT_15 = buttonOutputs.array[15];
    BUTTON_OUT_16 = buttonOutputs.array[16];
}

void UpdateServoOutputs(void)
{
    unsigned int i;
    signed int direction;
    signed long int range; //use long int to prevent overflow during math calculations
    signed long int range2;
    signed int sensorSignal;
    signed long int scaledSignal[NUM_SERVOS];
    signed long int scaledSignalOriginal[NUM_SERVOS]; //Make a copy so we don't overwrite the original when we mix signals
    signed long int output;
    signed int deadband;

    static signed int y1[4] = {0}; //variable for low pass filter
    static signed int y=0; //variable for low pass filter
    static signed int x=0; //variable for low pass filter

    static BOOL stickHoldToggleL = FALSE;
    static BOOL stickHoldToggleR = FALSE;
    static unsigned int inputHoldValue[NUM_SERVOS];
    unsigned int analogsBuffer[NUM_SERVOS];
    static unsigned int gyroSoftStartTimer;

    //GetAnalogs() returns a pointer to a volatile array, buffer it so the values don't change while we're in the middle of this function
    for (i=0; i<NUM_SERVOS; i++)
    {
        analogsBuffer[i] = GetAnalogs()->array[i];
    }

    //Some revisions of the controller requrie the gyro to be turned on manually.
    //Give it some time to power up or the servo gyro will whip to zero when first connected.
    if (GetPS3Report()[0] != 0xFF) //if the controller is connected and the gyro soft start countdown timer hasn't run out yet
    {
        if (gyroSoftStartTimer > 0)
        {
            gyroSoftStartTimer--;
            analogsBuffer[GYRO] = ServoSettings[GYRO].zeroPosition; //put in artificial value until the gyro has powered up.
        }
    }
    else
    {
        gyroSoftStartTimer = 30;
    }


    if (GetNewPress()->lStickPress == TRUE)
    {
        stickHoldToggleL = !stickHoldToggleL; //toggle servo hold on/off
    }
    if (GetNewPress()->rStickPress == TRUE)
    {
        stickHoldToggleR = !stickHoldToggleR; //toggle servo hold on/off
    }
    
    if (mecanumSteering >= MECANUM_REMAP_4) //override output with R_STICK_Y signal if we're remapping
    {
        analogsBuffer[mecanumSteering+2] = analogsBuffer[R_STICK_Y];
    }

    for (i=0; i<NUM_SERVOS; i++)
    {
        if ((ServoSettings[i].holdRecall == L_STICK && GetButtonPress()->psButton && GetNewPress()->lStickPress)||
            (ServoSettings[i].holdRecall == R_STICK && GetButtonPress()->psButton && GetNewPress()->rStickPress))
        {
            ServoSettings[i].servoRecallValue = (unsigned int)(servoState[i]/100); //store home position, divide by 100 to make unsigned long  fit unsigned int
            DataEEWrite(ServoSettings[i].servoRecallValue, EEPROM_SERVO_HOME + i);
        }
        //if the hold/recall function is on in relative mode, recall stored value to servostate
        else if ((ServoSettings[i].absOrRel == RELATIVE && ServoSettings[i].holdRecall == L_STICK && GetNewPress()->lStickPress)|| //recall home position
                 (ServoSettings[i].absOrRel == RELATIVE && ServoSettings[i].holdRecall == R_STICK && GetNewPress()->rStickPress))
        {
            ServoSettings[i].servoRecallValue = DataEERead(EEPROM_SERVO_HOME + i);
            servoState[i] = ((signed long)(ServoSettings[i].servoRecallValue))*100; //recall home position, multiply by 100 cancel out divide when stored
        }
        //if the hold/recall function is on in absolute mode, overwrite the stick value with stored value
        else if ((ServoSettings[i].absOrRel == ABSOLUTE && ServoSettings[i].holdRecall == R_STICK && stickHoldToggleR == TRUE)||
                 (ServoSettings[i].absOrRel == ABSOLUTE && ServoSettings[i].holdRecall == L_STICK && stickHoldToggleL == TRUE))
        {
            analogsBuffer[i] = inputHoldValue[i]; //overwrite stick value with hold value
        }
        else
        {
            inputHoldValue[i] = analogsBuffer[i]; //update hold value with stick value
        }
        //Channel-specific signal conditioning.
        switch (i)
        {
            case L_STICK_X:
            case L_STICK_Y:
            case R_STICK_X:
            case R_STICK_Y:
                sensorSignal = (signed int)analogsBuffer[i]-ServoSettings[i].zeroPosition;
                break;
            case L_TRIG:
            case R_TRIG:
                if (triggerLink == TRUE)
                {
                    sensorSignal = ((signed int)analogsBuffer[R_TRIG]-(signed int)analogsBuffer[L_TRIG])/2;
                }
                else
                {
                    sensorSignal = (signed int)analogsBuffer[i]-ServoSettings[i].zeroPosition-128;
                }
                break;
            case ACCEL_X:
            case ACCEL_Y:
            case ACCEL_Z:
            case GYRO:
                 //simple recursive low pass filter
                #define FILTER_COEF 3
                x = ((signed int)analogsBuffer[i] - ServoSettings[i].zeroPosition)*FILTER_COEF; //multiply, then divide to reduce rounding error
                y = y1[i-ACCEL_X] + (x - y1[i-ACCEL_X])/FILTER_COEF;
                y1[i-ACCEL_X] = y;
                sensorSignal = y/FILTER_COEF;
                break;
            default:
                sensorSignal = (signed int)analogsBuffer[i]-ServoSettings[i].zeroPosition;
                UART2PrintString("Error: case not handled.\n\r");
                break;
        }
         // special case for remapping r-stick y-axis when using mecanum mixing remapping
        if (mecanumSteering >= MECANUM_REMAP_4 && i == (mecanumSteering+2))
        {
            sensorSignal = (signed int)analogsBuffer[i]-ServoSettings[R_STICK_Y].zeroPosition;  //use the original home position calibration.
        }

        //Apply deadband
        deadband = (signed int)ServoSettings[i].deadband;
        if (sensorSignal > deadband)
        {
            sensorSignal = sensorSignal - deadband;
        }
        else if (sensorSignal < -deadband)
        {
            sensorSignal = sensorSignal + deadband;
        }
        else
        {
            sensorSignal = 0;
        }

        //Scale signal by sensitivity
        if (i == GYRO)
        {
            scaledSignal[i] = (signed long)ServoSettings[i].sensitivity*(signed long)sensorSignal/2 + ServoSettings[i].trim*10; //gyro is more sensitive, scale down
        }
        else
        {
            scaledSignal[i] = (signed long)ServoSettings[i].sensitivity*(signed long)sensorSignal + ServoSettings[i].trim*10; //beware of overflow
        }
        scaledSignalOriginal[i] = scaledSignal[i]; //we'll overwrite some signals if we mix, so make a copy
    }

    //////////////////////////////////////////////////////////
    //Split the loop to finish applying scalings to all the signals first, so that we can mix the outputs later with the scaled signals.

    for (i=0; i<NUM_SERVOS; i++)
    {
        //mix signals for Mecanum wheels
        if (mecanumSteering != MECANUM_OFF )
        {
            switch (i)
            {
                case L_STICK_X: //left front wheel
                    scaledSignal[L_STICK_X] = scaledSignalOriginal[L_STICK_Y] - scaledSignalOriginal[R_STICK_X] - scaledSignalOriginal[L_STICK_X];
                    break;
                case L_STICK_Y: //left rear wheel
                    scaledSignal[L_STICK_Y] = scaledSignalOriginal[L_STICK_Y] - scaledSignalOriginal[R_STICK_X] + scaledSignalOriginal[L_STICK_X];
                    break;
                case R_STICK_X: //right front wheel
                    scaledSignal[R_STICK_X] = scaledSignalOriginal[L_STICK_Y] + scaledSignalOriginal[R_STICK_X] + scaledSignalOriginal[L_STICK_X];
                    break;
                case R_STICK_Y: //right rear wheel
                    scaledSignal[R_STICK_Y] = scaledSignalOriginal[L_STICK_Y] + scaledSignalOriginal[R_STICK_X] - scaledSignalOriginal[L_STICK_X];
                    break;
                default:
                    break;
            }
        }
        else //if mecanum mixer is off
        {
            if (lStickDiffMix == TRUE) //check for diff steering
            {
                switch (i)
                {
                    case L_STICK_X:
                        scaledSignal[L_STICK_X] = scaledSignalOriginal[L_STICK_Y] - scaledSignalOriginal[L_STICK_X]; //left wheel channel
                        break;
                    case L_STICK_Y:
                        scaledSignal[L_STICK_Y] = scaledSignalOriginal[L_STICK_Y] + scaledSignalOriginal[L_STICK_X]; //right wheel channel
                        break;
                    default:
                        break;
                }
            }
            if (rStickDiffMix == TRUE) //check for diff steering
            {
                switch (i)
                {
                    case R_STICK_X:
                        scaledSignal[R_STICK_X] = scaledSignalOriginal[R_STICK_Y] - scaledSignalOriginal[R_STICK_X]; //left wheel channel
                        break;
                    case R_STICK_Y:
                        scaledSignal[R_STICK_Y] = scaledSignalOriginal[R_STICK_Y] + scaledSignalOriginal[R_STICK_X]; //right wheel channel
                        break;
                    default:
                        break;
                }
            }
        }

//*******************CHECK BOUNDARIES AND WRITE OUTPUTS*****************************//
        //Check direction
        if (ServoSettings[i].direction == FORWARD)
        {
            direction = 1;
        }
        else
        {
            direction = -1;
        }

        //////////////SERVO OUTPUT////////////////////////////////////
        if (ServoSettings[i].servoOrMotor == SERVO)
        {
            if (*OCxCON1_REG_PTRS[i] != 0x0006)
            {
                *OCxCON1_REG_PTRS[i] = 0x0006; //set source as TMR2
            }
            if (i == GYRO)
            {
                T3CONbits.TCKPS = 0b01; //set presacel on TMR3 to 1:64 (for PWM10)
            }

            *OCxRS_REG_PTRS[i] = ServoSettings[i].servoPeriodMs*2000; //set period, 10 to 20 ms

            //Servo Position Mode
            if (ServoSettings[i].absOrRel == ABSOLUTE)
            {
                //check range of motion
                range = ServoSettings[i].range*10;
                if ( scaledSignal[i] > range)
                {
                    scaledSignal[i] = range;
                }
                else if (scaledSignal[i] < -range)
                {
                    scaledSignal[i] = -range;
                }
                //check stops
                output = 3000+ServoSettings[i].offset*10 + direction*scaledSignal[i]; //output, 0.5uS per count
                if ( output > STOP_HIGH)
                {
                    output = STOP_HIGH;
                }
                else if (output < STOP_LOW)
                {
                    output = STOP_LOW;
                }

                //write output
                *OCxR_REG_PTRS[i] = output;
            }

            /////////////////SERVO RELATIVE MODE/////////////////
            else if (ServoSettings[i].absOrRel == RELATIVE)
            {
                //scale to normalize sensitivity.  Multiply everything by 100, then divide by 100 counter discretization noise when scaling.
                servoState[i]=servoState[i]+direction*scaledSignal[i];

                range = 300000 + 1000*(signed long)(ServoSettings[i].offset + ServoSettings[i].range); //set stops, beware of overflow!
                range2 = 300000 + 1000*(signed long)(ServoSettings[i].offset - ServoSettings[i].range); //set stops

                if (servoState[i] > range)
                {
                    servoState[i] = range;
                }
                else if (servoState[i] < range2)
                {
                    servoState[i] = range2;
                }
                //write output
                if ( servoState[i] > (signed long)STOP_HIGH*100)
                {
                    servoState[i] = (signed long)STOP_HIGH*100;
                }
                else if (servoState[i] < (signed long)STOP_LOW*100)
                {
                    servoState[i] = (signed long)STOP_LOW*100;
                }
                *OCxR_REG_PTRS[i] = servoState[i]/100;
            }
        }

        /////////////////MOTOR PWM MODE/////////////////
        else if (ServoSettings[i].servoOrMotor == MOTOR)
        {
            unsigned long int pwmPeriod;
            if (*OCxCON1_REG_PTRS[i] != 0x1C06) //0x1C06 == 0b0001110000000110 (use system clock, edge aligned PWM mode.
            {
                *OCxCON1_REG_PTRS[i] = 0x1C06;
            }
            if (i == GYRO)
            {
                T3CONbits.TCKPS = 0b00; //set presacel on TMR3 to 1:1 (for PWM10)
            }
            
            pwmPeriod = GetInstructionClock()/(ServoSettings[i].pwmFrequency*1000);
            *OCxRS_REG_PTRS[i] = pwmPeriod;

            //calculate duty cycle assuming 20kHz, scale it later
            ///////////////MOTOR ABSOLUTE/////////////
            if (ServoSettings[i].absOrRel == ABSOLUTE)
            {
                range = ServoSettings[i].range*10;

                if ( scaledSignal[i] > range)
                {
                    scaledSignal[i] = range;
                }
                else if (scaledSignal[i] < -range)
                {
                    scaledSignal[i] = -range;
                }
                output = 10000+ServoSettings[i].offset*100 + 5*direction*scaledSignal[i]; //scale signal to normalize sensitivity

                if ( output > 20000)
                {
                    output = 20000;
                }
                else if (output < 0)
                {
                    output = 0;
                }
                *OCxR_REG_PTRS[i] = (output*pwmPeriod)/20000;
            }

            ///////////////MOTOR RELATIVE/////////////
            else if (ServoSettings[i].absOrRel == RELATIVE)
            {
                servoState[i]=servoState[i]+direction*scaledSignal[i]/20; //scale signal to normalize sensitivity
                range = 10000 + 50*(signed long)(ServoSettings[i].offset + ServoSettings[i].range); //set stops, beware of overflow!
                range2 = 10000 + 50*(signed long)(ServoSettings[i].offset - ServoSettings[i].range); //set stops

                if (servoState[i] > range)
                {
                    servoState[i] = range;
                }
                else if (servoState[i] < range2)
                {
                    servoState[i] = range2;
                }

                //write output
                if ( servoState[i] > 20000)
                {
                    servoState[i] = 20000;
                }
                else if (servoState[i] < 0)
                {
                    servoState[i] = 0;
                }

                *OCxR_REG_PTRS[i] = servoState[i]*pwmPeriod/20000;
            } //end motor relative if
        } //end motor PWM if
    } //end NUM_SERVO for loop
}

void ResetOutputs(CONTROLLER_IN* ControllerIn)
{
    UART2PrintString("***Reset Outputs.***\n\r");
    ControllerIn->select = 0;
    ControllerIn->lStickPress = 0;
    ControllerIn->rStickPress = 0;
    ControllerIn->start = 0;
    ControllerIn->dPadUp = 0;
    ControllerIn->dPadRight = 0;
    ControllerIn->dPadDown = 0;
    ControllerIn->dPadLeft = 0;
    ControllerIn->lTriggerPress = 0;
    ControllerIn->rTriggerPress = 0;
    ControllerIn->lBumper = 0;
    ControllerIn->rBumper = 0;
    ControllerIn->triangle = 0;
    ControllerIn->circle = 0;
    ControllerIn->cross = 0;
    ControllerIn->square = 0;
    ControllerIn->psButton = 0;

    ControllerIn->lStickX = ServoSettings[0].zeroPosition;
    ControllerIn->lStickY = ServoSettings[1].zeroPosition;
    ControllerIn->rStickX = ServoSettings[2].zeroPosition;
    ControllerIn->rStickY = ServoSettings[3].zeroPosition;
    ControllerIn->lTriggerAnalog = ServoSettings[4].zeroPosition;
    ControllerIn->rTriggerAnalog = ServoSettings[5].zeroPosition;
    ControllerIn->accelX_1 = ServoSettings[6].zeroPosition/256;
    ControllerIn->accelX_0 = ServoSettings[6].zeroPosition;
    ControllerIn->accelY_1 = ServoSettings[7].zeroPosition/256;
    ControllerIn->accelY_0 = ServoSettings[7].zeroPosition;
    ControllerIn->accelZ_1 = ServoSettings[8].zeroPosition/256;
    ControllerIn->accelZ_0 = ServoSettings[8].zeroPosition;
    ControllerIn->gyroZ_1 = ServoSettings[9].zeroPosition/256;
    ControllerIn->gyroZ_0 = ServoSettings[9].zeroPosition;

    UpdateCurrentStateBuffer(ControllerIn); //update controller input fields
    SpiInputOverride(); //check if we're overriding inputs with the
    UpdateNewPress();
    if (disconnectToggleReset == 1)
    {
        UpdateButtonOutputs(2,0); //2: config mode not active, but re-initialize the button toggle states
    }
    else
    {
        UpdateButtonOutputs(0,0); //0: config mode not active
    }
    UpdateServoOutputs();
}

/*******************************************************************************
Function: GetSpiBusMode()

Precondition:
 None

Overview:
    This subroutine returns the SPI bus mode

Input:

Output:
0,1,2,3

*******************************************************************************/
unsigned int GetSpiBusMode(void)
{
    return spiBusMode;
}

/*******************************************************************************
Function: PrintIndexLabel(index)

Precondition:
 None

Overview:
    This subroutine prints the controller input associated with the channel 
number.

Input: Output
 * //////////////SERVOS///////////////
 * 0: Left Stick x-axis
 * 1: Left Stick y-axis
 * 2: Right Stick x-axis
 * 3: Right Stick y-axis
 * 4: Left Trigger Analog
 * 5: Right Trigger Analog
 * 6: X-accel
 * 7: Y-accel
 * 8: Z-accel
 * 9: Z-gyro
 * //////////////BUTTONS///////////////
 * 10: select
 * 11: lStickPress
 * 12: rStickPress
 * 13: start
 * 14: dPadUp
 * 15: dPadRight
 * 16: dPadDown
 * 17: dPadLeft
 * 18: lTriggerPress
 * 19: rTriggerPress
 * 20: lBumper
 * 21: rBumper
 * 22: triangle
 * 23: circle
 * 24: cross
 * 25: square
 * 26: psButton

*******************************************************************************/

void PrintIndexLabel(unsigned int index)
{
    char* indexLabels[] = {\
    "L-Stick X-axis",\
    "L-Stick Y-axis",\
    "R-Stick X-axis",\
    "R-Stick Y-axis",\
    "L-Trig analog",\
    "R-Trig analog",\
    "X-Accelerometer",\
    "Y-Accelerometer",\
    "Z-Accelerometer",\
    "Z-Gyroscope",\
    "Select button",\
    "L-Stick button",\
    "R-Stick button",\
    "Start button",\
    "D-Pad Up",\
    "D-Pad Right",\
    "D-Pad Down",\
    "D-Pad Left",\
    "L-Trig button",\
    "R-Trig button",\
    "L-Bumper",\
    "R-Bumper",\
    "Triangle button",\
    "Circle button",\
    "Cross button",\
    "Square button",\
    "PS button"
    };
    if (index > 26)
    {
        UART2PrintString("Error: Index out of range. \n\r");
    }
    else
    {
        UART2PrintString("Channel ");
        UART2PutDecInt(index);
        UART2PutChar(':');
        UART2PutChar(' ');
        UART2PrintString(indexLabels[index]); //Careful, don't try to call a label that doesn't exist!
        UART2PrintString("\n\r");
    }
}

/*******************************************************************************
Function: _T3Interrupt

Precondition:
 Triggers when T3 timer elapses

Overview:
 This ISR is used to generate a PWM signal for PWM10

Input:
 configFlag == TRUE will continue to run the routine.
 configFlag == FALSE will exit the routine.

Output:
  0 = save and exited;
  1 = stay in config mode;

*******************************************************************************/
////////////////////ISR for T3 INTERRUPT (PWM10 channel, for gyro)////////////////////
void __attribute__((__interrupt__, auto_psv))_T3Interrupt(void)
{
    if (PWM10_OUT == 0)
    {
        PWM10_OUT = 1;
        PR3 = OC10R+1; //new period registor for next toggle
    }
    else
    {
        PWM10_OUT = 0;
        PR3 = OC10RS-OC10R+1;
    }
    TMR3 = 0; //reset TMR3
    _T3IF = 0; //clear interrupt flag
}



